from setuptools import setup
import os.path


PACKAGE_NAME = 'ruslanio'


def get_full_path(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(here, rel_path)


def read(rel_path):
    with open(get_full_path(rel_path), 'r') as f:
        return f.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1].strip()
    else:
        raise RuntimeError("Unable to find version string.")

def get_packages(rel_path):
    path = get_full_path(rel_path)
    result = []
    if os.path.isdir(path):
        names = os.listdir(path)
        if '__init__.py' in names:
            result.append(rel_path)
        for name in names:
            new_path = os.path.join(rel_path, name)
            result.extend(get_packages(new_path))
    return result


setup(
    name=PACKAGE_NAME,
    version=get_version(f'{PACKAGE_NAME}/__init__.py'),
    description='My personal tools for my projects',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Ruslan Baynazarov',
    author_email='',
    url='https://github.com/hocop/ruslanio',
    packages=get_packages(PACKAGE_NAME),
    license="MIT",
    include_package_data=True,
    install_requires=read('requirements.txt').splitlines()
)



import asyncio
import typing as tp
from collections import OrderedDict
import json
import datetime


class LRU:
    def __init__(self, maxsize=128, class_method=False, logger=None):
        """
        :param maxsize: Use maxsize as None for unlimited size cache
        :param class_method: Set True to ignore the first "self" argument
        :param logger: if set, this function will print debug logs when cache is used
        """
        self.lru = _LRU(maxsize=maxsize)
        self.class_method = class_method
        self.logger = logger

    def __call__(self, func):
        def wrapper(*args, use_cache=True, **kwargs):
            if use_cache:
                # Make key for caching
                key_args = args[1:] if self.class_method else args
                key = _KEY(key_args, kwargs)
                # Call function or use cached value
                if not key in self.lru:
                    self.lru[key] = func(*args, **kwargs)
                else:
                    if self.logger is not None:
                        self.logger.debug(f'Using cached {key_args} {kwargs}')
                return self.lru[key]
            else:
                return func(*args, **kwargs)

        wrapper.__name__ += func.__name__

        return wrapper


class AsyncLRU:
    def __init__(self, maxsize=128, class_method=False, logger=None):
        """
        :param maxsize: Use maxsize as None for unlimited size cache
        :param class_method: Set True to ignore the first "self" argument
        :param logger: if set, this function will print debug logs when cache is used
        """
        self.lru = _LRU(maxsize=maxsize)
        self.class_method = class_method
        self.logger = logger

    def __call__(self, func):
        async def wrapper(*args, use_cache=True, **kwargs):
            if use_cache:
                # Make key for caching
                key_args = args[1:] if self.class_method else args
                key = _KEY(key_args, kwargs)
                # Call function or use cached value
                if not key in self.lru:
                    self.lru[key] = asyncio.create_task(func(*args, **kwargs))
                else:
                    if self.logger is not None:
                        self.logger.debug(f'Using cached {key_args} {kwargs}')
                return await self.lru[key]
            else:
                return await func(*args, **kwargs)

        wrapper.__name__ += func.__name__

        return wrapper


class RedisAsyncLRU:
    def __init__(self, name, expire: datetime.timedelta | None = None, class_method=False, logger=None):
        """
        :param name: Name of function, for cache prefix
        :param expire: Expiration time
        :param class_method: Set True to ignore the first "self" argument
        :param logger: if set, this function will print debug logs when cache is used
        """
        self.name = name
        self.class_method = class_method
        self.lru = _RedisDict(expire=expire)
        self.logger = logger

    def __call__(self, func):
        async def wrapper(*args, use_cache=True, **kwargs):
            if use_cache:
                # Make key for caching
                key_args = args[1:] if self.class_method else args
                key = self.name + ' -- ' + str(_KEY(key_args, kwargs))
                # Call function or use cached value
                if not key in self.lru:
                    self.lru[key] = await func(*args, **kwargs)
                else:
                    if self.logger is not None:
                        self.logger.debug(f'Using cached {self.name} {key_args} {kwargs}')
                return self.lru[key]
            else:
                return await func(*args, **kwargs)

        wrapper.__name__ += func.__name__

        return wrapper


class _RedisDict:
    def __init__(self, expire):
        import redis
        self.redis = redis.Redis()
        self.expire = expire

    def __getitem__(self, key: str):
        value = self.redis.get(key)
        value = json.loads(value)
        return value

    def __contains__(self, key: str):
        return key in self.redis

    def __setitem__(self, key, value):
        value = json.dumps(value, ensure_ascii=False)
        self.redis.set(key, value, ex=self.expire)


class _LRU(OrderedDict):
    def __init__(self, maxsize, *args, **kwargs):
        self.maxsize = maxsize
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self.maxsize and len(self) > self.maxsize:
            oldest = next(iter(self))
            del self[oldest]


class _KEY:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        kwargs.pop("use_cache", None)

    def __eq__(self, obj):
        return hash(self) == hash(obj)

    def _get_immutable(self):
        def _hash(param: tp.Any):
            if isinstance(param, tuple):
                return tuple(map(_hash, param))
            if isinstance(param, dict):
                return tuple(map(_hash, param.items()))
            elif hasattr(param, "__dict__"):
                return str(vars(param))
            else:
                return str(param)

        return _hash(self.args) + _hash(self.kwargs)

    def __str__(self) -> str:
        return str(self._get_immutable())

    def __hash__(self):
        return hash(self._get_immutable())
